#coding:utf-8
import os
import gzip
import io
import hashlib
import subprocess

from six import text_type, BytesIO

from cactus.utils.helpers import checksum


class FakeTime:
    """
    Monkey-patch gzip.time to avoid changing files every time we deploy them.
    """
    def time(self):
        return 1111111111.111


def compressString(s):
    """Gzip a given string."""

    gzip.time = FakeTime()

    zbuf = BytesIO()
    zfile = gzip.GzipFile(mode='wb', compresslevel=9, fileobj=zbuf)
    zfile.write(s)
    zfile.close()
    return zbuf.getvalue()


def fileSize(num):
    for x in ['b', 'kb', 'mb', 'gb', 'tb']:
        if num < 1024.0:
            return "%.0f%s" % (num, x)
        num /= 1024.0

def calculate_file_checksum(path):
    """
    Calculate the MD5 sum for a file:
    Read chunks of a file and update the hasher.
    Returns the hex digest of the md5 hash.
    """
    hasher = hashlib.md5()
    with io.FileIO(path, 'r') as fp:
        while True:
            buf = fp.read(65536)
            if not buf:
                break
            hasher.update(buf)
    return hasher.hexdigest()

def file_changed_hash(path):
    info = os.stat(path)
    hashKey = text_type(info.st_mtime) + text_type(info.st_size)
    return checksum(hashKey.encode('utf-8'))



import types
import logging

from boto.route53.connection import Route53Connection
from boto.route53.record import ResourceRecordSets

from boto.s3.connection import S3Connection
from boto.s3.website import RedirectLocation

# Generic steps for custom domains can be found here
# http://docs.aws.amazon.com/AmazonS3/latest/dev/website-hosting-custom-domain-walkthrough.html

# Hosted zone tables

# https://forums.aws.amazon.com/thread.jspa?threadID=116724
# http://docs.aws.amazon.com/general/latest/gr/rande.html#s3_region

# Amazon, you owe me four hours of my life...

HOSTED_ZONES = {
    # US Standard
    "s3-website-us-east-1.amazonaws.com": "Z3AQBSTGFYJSTF",
    # US West (Oregon) Region
    "s3-website-us-west-2.amazonaws.com": "Z3BJ6K6RIION7M",
    # US West (Northern California) Region
    "s3-website-us-west-1.amazonaws.com": "Z2F56UZL2M1ACD",
    # EU (Ireland) Region
    "s3-website-eu-west-1.amazonaws.com": "Z1BKCTXD74EZPE",
    # Asia Pacific (Singapore) Region
    "s3-website-ap-southeast-1.amazonaws.com": "Z3O0J2DXBE1FTB",
    # Asia Pacific (Sydney) Region
    "s3-website-ap-southeast-2.amazonaws.com": "Z1WCIGYICN2BYD",
    # Asia Pacific (Tokyo) Region
    "s3-website-ap-northeast-1.amazonaws.com": "Z2M4EHUR26P7ZW",
    # South America (Sao Paulo) Region
    "s3-website-sa-east-1.amazonaws.com": "Z7KQH4QJS55SO",
    # AWS GovCloud (US)
    "s3-website-us-gov-west-1.amazonaws.com": "Z31GFT0UA1I2HV",
}


class AWSBucket(object):

    def __init__(self, accessKey, secretKey, name):

        self.name = name

        self.accessKey = accessKey
        self.secretKey = secretKey

        self.connection = S3Connection(
            aws_access_key_id=accessKey,
            aws_secret_access_key=secretKey,
        )

        self._cache = {}

    def bucket(self):
        try:
            return self.connection.get_bucket(self.name)
        except:
            return None

    def create(self):
        logging.info('Create bucket %s', self.name)
        self.connection.create_bucket(self.name, policy='public-read')

    def isCreated(self):
        return self.bucket() is not None

    def configureWebsite(self):
        logging.info('Configuring website endpoint %s', self.name)
        self.bucket().configure_website('index.html', 'error.html')

    def configureRedirect(self, url):
        logging.info('Setup redirect %s -> %s', self.name, url)
        self.bucket().configure_website(
            redirect_all_requests_to=RedirectLocation(hostname=url))

    def websiteEndpoint(self):
        return self.bucket().get_website_endpoint()




class AWSDomain(object):

    def __init__(self, accessKey, secretKey, domain):

        self.domain = domain

        self.accessKey = accessKey
        self.secretKey = secretKey

        self.connection = Route53Connection(
            aws_access_key_id=accessKey,
            aws_secret_access_key=secretKey,
        )

        self._cache = {}

    @property
    def id(self):
        return self.hostedZone()['HostedZone']['Id'].replace('/hostedzone/', '')

    @property
    def fullDomain(self):
        parts = self.domain.split(".")
        return parts[len(parts) - 2] + "." + parts[len(parts) - 1]

    @property
    def dnsDomain(self):
        return self.domain + "."

    def isValidDomain(self):
        pass

    def isNakedDomain(self):
        pass

    def records(self):
        pass

    def createHostedZone(self):

        logging.info('Creating hosted zone for %s', self.fullDomain)

        self.connection.create_hosted_zone(self.fullDomain)

    def hostedZone(self):
        if not "hostedZone" in self._cache:
            hostedZone = self.connection.get_hosted_zone_by_name(self.fullDomain)
            if not hostedZone:
                return
            self._cache["hostedZone"] = hostedZone.get("GetHostedZoneResponse", None)
        return self._cache["hostedZone"]

    def nameServers(self):
        hostedZone = self.hostedZone()

        if hostedZone:
            return self.hostedZone()["DelegationSet"]["NameServers"]
        else:
            return None

    def records(self):
        return self.connection.get_all_rrsets(self.id)

    def createRecord(self, name, recordType, values, ttl=60*60*3):
        self._changeRecord("CREATE", name, recordType, values, ttl)

    def deleteRecord(self, name, recordType, values, ttl=60*60*3):
        self._changeRecord("DELETE", name, recordType, values, ttl)

    def _changeRecord(self, change, name, recordType, values, ttl):

        logging.info('%s record %s:%s in zone %s', change, name, recordType, self.domain)

        if type(values) is not types.ListType:
            values = [values]

        changes = ResourceRecordSets(self.connection, self.id)
        change = changes.add_change(change, name, recordType, ttl)

        for value in values:
            change.add_value(value)

        changes.commit()


    def createAlias(self, name, recordType, aliasHostedZoneId, aliasDNSName, identifier=None, weight=None, comment=""):
        self._changeAlias("CREATE", name, recordType, aliasHostedZoneId, aliasDNSName, identifier, weight, comment)

    def deleteAlias(self, name, recordType, aliasHostedZoneId, aliasDNSName, identifier=None, weight=None, comment=""):
        self._changeAlias("DELETE", name, recordType, aliasHostedZoneId, aliasDNSName, identifier, weight, comment)

    def _changeAlias(self, change, name, recordType, aliasHostedZoneId, aliasDNSName, identifier, weight, comment):

        logging.info('%s alias %s:%s in zone %s', change, name, recordType, self.domain)

        changes = ResourceRecordSets(self.connection, self.id, comment)
        change = changes.add_change(change, name, recordType, identifier=identifier, weight=weight)
        change.set_alias(aliasHostedZoneId, aliasDNSName)
        changes.commit()


    def delete(self, record):

        if record.alias_dns_name:
            self.deleteAlias(record.name, record.type,
                record.alias_hosted_zone_id, record.alias_dns_name,
                identifier=record.identifier, weight=record.weight)
        else:
            self.deleteRecord(record.name, record.type, record.resource_records,
                ttl=record.ttl)

    def pointRootToBucket(self):

        # Make sure the correct bucket exists and is ours
        bucket = AWSBucket(self.accessKey, self.secretKey, self.domain)
        endpoint = bucket.websiteEndpoint()
        endpointDomain = endpoint[len(self.dnsDomain):]

        # Remove old A record for the root domain
        for record in self.records():
            if record.type in ["A", "CNAME"] and record.name == self.dnsDomain:
                self.delete(record)

        # Create new root domain record that points to the bucket
        self.createAlias(self.dnsDomain, "A", HOSTED_ZONES[endpointDomain], endpointDomain)


    def setupRedirect(self):

        redirectDomain = "www.%s" % self.domain
        redirectDNSDomain = redirectDomain + '.'

        bucket = AWSBucket(self.accessKey, self.secretKey, redirectDomain)

        if bucket.isCreated():
            logging.info("Bucket with name %s already exists, so skipping redirect bucket setup. \
                If you've set this up before, this will still work. Delete the bucket if you want Cactus to\
                set it up again.", redirectDomain)
            return

        bucket.create()
        bucket.configureRedirect(self.domain)

        for record in self.records():
            if record.type == "CNAME" and record.name == redirectDNSDomain:
                self.delete(record)

        self.createRecord(redirectDomain, "CNAME", [self.domain])

    def setup(self):

        if not self.hostedZone():
            self.createHostedZone()

        self.pointRootToBucket()

        if not self.domain.startswith("www."):
            self.setupRedirect()



#coding:utf-8

import pyrax

from cactus.deployment.engine import BaseDeploymentEngine
from cactus.deployment.cloudfiles.auth import CloudFilesCredentialsManager
from cactus.deployment.cloudfiles.file import CloudFilesFile


class CloudFilesDeploymentEngine(BaseDeploymentEngine):
    CredentialsManagerClass = CloudFilesCredentialsManager
    FileClass = CloudFilesFile

    config_bucket_name = "cloudfiles-bucket-name"
    config_bucket_website = "cloudfiles-bucket-website"

    def _create_connection(self):
        username, api_key = self.credentials_manager.get_credentials()
        pyrax.set_setting("identity_type", "rackspace")
        pyrax.set_credentials(username, api_key)
        return pyrax.connect_to_cloudfiles()

    def get_bucket(self):
        try:
            return self.get_connection().get_container(self.bucket_name)
        except pyrax.exceptions.NoSuchContainer:
            return None

    def create_bucket(self):
        #TODO: Handle errors
        conn = self.get_connection()
        container = conn.create_container(self.bucket_name)
        container.set_web_index_page(self._index_page)
        container.set_web_error_page(self._error_page)
        container.make_public()
        return container

    def get_website_endpoint(self):
        return self.bucket.cdn_uri



#coding:utf-8
import functools

from cactus.utils.internal import getargspec
from cactus.plugin import defaults


class PluginManager(object):
    def __init__(self, site, loaders):
        self.site = site
        self.loaders = loaders
        self.reload()

        for plugin_method in defaults.DEFAULTS:
            if not hasattr(self, plugin_method):
                setattr(self, plugin_method, functools.partial(self.call, plugin_method))

    def reload(self):
        plugins = []
        for loader in self.loaders:
            plugins.extend(loader.load())

        self.plugins = sorted(plugins, key=lambda plugin: plugin.ORDER)

    def call(self, method, *args, **kwargs):
        """
        Call each plugin
        """
        for plugin in self.plugins:
            _meth = getattr(plugin, method)
            _meth(*args, **kwargs)

    def preBuildPage(self, site, page, context, data):
        """
        Special call as we have changed the API for this.

        We have two calling conventions:
        - The new one, which passes page, context, data
        - The deprecated one, which also passes the site (Now accessible via the page)
        """
        for plugin in self.plugins:
            # Find the correct calling convention
            new = [page, context, data]
            deprecated = [site, page, context, data]
            arg_lists = dict((len(l), l) for l in [deprecated, new])

            try:
                # Try to find the best calling convention
                n_args = len(getargspec(plugin.preBuildPage).args)
                # Just use the new calling convention if there's fancy usage of
                # *args, **kwargs that we can't control.
                arg_list = arg_lists.get(n_args, new)
            except NotImplementedError:
                # If we can't get the number of args, use the new one.
                arg_list = new

            # Call with the best calling convention we have.
            # If that doesn't work, then we'll let the error escalate.
            context, data = plugin.preBuildPage(*arg_list)

        return context, data



#coding:utf-8

class CacheDurationPlugin(object):
    """
    A plugin to make the default cache expiry configurable via a "cache-duration" configuration setting
    """

    def preDeploy(self, site):
        """
        Load the cache duration from the config
        """
        self.cache_duration = site.config.get("cache-duration")

    def preDeployFile(self, file):
        """
        Set the cache duration expiry on the file
        """
        if self.cache_duration is not None and not file.is_fingerprinted:
            file.cache_control = self.cache_duration



import os
import sys
import pipes
import shutil
import subprocess

"""
This plugin uses glue to sprite images:
http://glue.readthedocs.org/en/latest/quickstart.html

Install:

(Only if you want to sprite jpg too)
brew install libjpeg

(Only if you want to optimize pngs with optipng)
brew install optipng

sudo easy_install pip
sudo pip uninstall pil
sudo pip install pil
sudo pip install glue
"""

try:
    import glue
except Exception as e:
    sys.exit('Could not use glue: %s\nMaybe install: sudo easy_install glue' % e)


IMG_PATH = 'static/img/sprites'
CSS_PATH = 'static/css/sprites'

KEY = '_PREV_CHECKSUM'

def checksum(path):
    command = 'md5 `find %s -type f`' % pipes.quote(IMG_PATH)
    return subprocess.check_output(command, shell=True)

def preBuild(site):

    currChecksum = checksum(IMG_PATH)
    prevChecksum = getattr(site, KEY, None)

    # Don't run if none of the images has changed
    if currChecksum == prevChecksum:
        return

    if os.path.isdir(CSS_PATH):
        shutil.rmtree(CSS_PATH)

    os.mkdir(CSS_PATH)
    os.system('glue --cachebuster --crop --optipng "%s" "%s" --project' % (IMG_PATH, CSS_PATH))

    setattr(site, KEY, currChecksum)



#coding:utf-8
import os
import shutil
from cactus.static.external.exceptions import ExternalFailure

from cactus.static.external import External
from cactus.tests import SiteTestCase


class TestExternal(External):
    runs = []

    def run(self):
        out = super(TestExternal, self).run()
        if self.accepted():
            TestExternal.runs.append(self)
        return out


class DummyCriticalFailingProc(TestExternal):
    supported_extensions = ('src',)
    output_extension = 'dst'
    critical = True

    def _run(self):
        raise OSError('Error.')


class DummyOptionalFailingProc(TestExternal):
    supported_extensions = ('src',)
    output_extension = 'dst'
    critical = False

    def _run(self):
        raise OSError('Error.')


class DummyExternal(TestExternal):
    supported_extensions = ('src',)
    critical = False

    def _run(self):
        shutil.move(self.src, self.dst)


class ExtensionDummyProc(DummyExternal):
    supported_extensions = ('src',)
    output_extension = 'blah'


class DummyProc(DummyExternal):
    output_extension = 'dst'


class DummyOptimizer(DummyExternal):
    supported_extensions = ('dst',)
    output_extension = 'dst'


class SourceOptimizer(DummyExternal):
    supported_extensions = ('src',)
    output_extension =  'src'


class UnrelatedOptimizer(DummyExternal):
    supported_extensions = ('aaa',)
    output_extension = 'bbb'


class DiscardingProcessor(DummyProc):
    def _run(self):
        self.discard()



class TestStaticExternals(SiteTestCase):
    """
    Test that externals are called properly, and that exceptions are handled properly.
    """
    def get_config_for_test(self):
        return {"optimize": ["src", "dst"]}

    def setUp(self):
        super(TestStaticExternals, self).setUp()
        self.site.external_manager.clear()

        # Write an empty file
        self.dummy_static = 'test.src'
        open(os.path.join(self.site.static_path, self.dummy_static), 'wb')

        TestExternal.runs = []

    def test_critical(self):
        """
        Test that failures on critical processors are escalated
        """
        self.site.external_manager.register_processor(DummyCriticalFailingProc)
        self.assertRaises(ExternalFailure, self.site.build)

    def test_non_critical(self):
        """
        Test that failures on non-critical processors are ignored
        """
        self.site.external_manager.register_processor(DummyOptionalFailingProc)
        self.site.build()

    def test_deregister(self):
        """
        Test that we can de-register a processor.
        """
        self.site.external_manager.register_processor(DummyCriticalFailingProc)
        self.assertRaises(ExternalFailure, self.site.build)
        self.site.external_manager.deregister_processor(DummyCriticalFailingProc)
        self.site.build()

    def test_run(self):
        """
        Test that processors and optimizers run
        """
        self.site.external_manager.register_processor(DummyProc)
        self.site.external_manager.register_optimizer(DummyOptimizer)
        self.site.external_manager.register_optimizer(UnrelatedOptimizer)
        self.site.build()

        self.assertEqual(2, len(TestExternal.runs))
        proc, opti = TestExternal.runs

        self.assertIsInstance(proc, DummyProc)
        self.assertIsInstance(opti, DummyOptimizer)

    def test_extensions(self):
        """
        Test that processors extensions are taken into account
        """
        self.site.external_manager.register_processor(ExtensionDummyProc)
        self.site.build()

        for static in self.site.static():
            if static.src_filename == self.dummy_static:
                self.assertEqual(static.final_extension, ExtensionDummyProc.output_extension)
                break
        else:
            self.fail("Did not find {0}".format(self.dummy_static))

    def test_discard(self):
        """
        Test that we discard files properly
        """
        self.site.external_manager.register_processor(DiscardingProcessor)
        self.site.external_manager.register_optimizer(SourceOptimizer)

        self.site.build()

        self.assertEqual(0, len(TestExternal.runs))  # Check that our optimizer does not run
        self.assertFileDoesNotExist(os.path.join(self.site.build_path, "static", "test.dst"))
        self.assertFileDoesNotExist(os.path.join(self.site.build_path, "static", "test.src"))

        for static in self.site.static():
            if static.src_filename == self.dummy_static:
                self.assertTrue(static.discarded)



#coding:utf-8
import os
from cactus.tests import SiteTestCase


class TestSiteContext(SiteTestCase):
    """
    Test that the proper context is provided to the pages

    Includes the built-in site context ('CACTUS'), and custom context.
    """
    def setUp(self):
        super(TestSiteContext, self).setUp()

        page = "{{ a }}\n{{ b }}"
        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write(page)

        self.site.build()

    def get_config_for_test(self):
        return {"context": {"a":"1", "b":True}}

    def test_site_context(self):
        """
        Test that the site context is provided.
        """
        self.assertEqual(
            sorted([page.source_path for page in self.site.context()['CACTUS']['pages']]),
            sorted(['error.html', 'index.html', 'test.html'])
        )

    def test_custom_context(self):
        """
        Test that custom context is provided
        """
        with open(os.path.join(self.site.build_path, "test.html")) as f:
            a, b = f.read().split("\n")

        self.assertEqual(a, "1")
        self.assertEqual(b, "True")

class TestCustomPageContext(SiteTestCase):
    """
    Test that custom context in the header of pages is feeded to a page

    Includes the built-in site context ('CACTUS'), and custom context.
    """
    def setUp(self):
        super(TestCustomPageContext, self).setUp()

        page = "a: 1\n\tb: Monkey\n{{ a }}\n{{ b }}"
        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write(page)

        self.site.build()

    def test_custom_context(self):
        """
        Test that custom context is provided
        """
        with open(os.path.join(self.site.build_path, "test.html")) as f:
            a, b = f.read().split("\n")

        self.assertEqual(a, "1")
        self.assertEqual(b, "Monkey")



#coding:utf-8
from django.core.management.commands.makemessages import Command as MakeMessagesCommand
from django.core.management.commands.compilemessages import Command as CompileMessagesCommand

from cactus.utils.filesystem import chdir

DEFAULT_COMMAND_KWARGS = {
    # Command Options
    "verbosity": 3,
    "settings": None,
    "pythonpath": None,
    "traceback": True,
    "all": False,
    'force_color': False,
    'no_color': True,
    'exclude': [],
    'add_location': False,
    'ignore_patterns': [],
    'fuzzy': False,
}

DEFAULT_MAKEMESSAGES_KWARGS = {
    # MakeMessages Options: Default
    "domain": "django",
    "extensions": [],
    "symlinks": False,
    "use_default_ignore_patterns": True,
    "no_wrap": False,
    "no_location": False,
    "no_obsolete": False,
    "keep_pot": False
}

def WrappedCommandFactory(wrapped, default_kwargs=None):
    # Compose a list of kwargs for future runs
    base_kwargs = {}
    base_kwargs.update(DEFAULT_COMMAND_KWARGS)
    if default_kwargs is not None:
        base_kwargs.update(default_kwargs)

    class WrappedCommand(object):
        """
        Wraps a Django management command
        """
        def __init__(self, site):
            self.site = site

        def execute(self):
            kwargs = {"locale": self.site.locale}
            kwargs.update(base_kwargs)

            cmd = wrapped()
            with chdir(self.site.path):
                cmd.execute(**kwargs)  # May raise an exception depending on gettext install.

    return WrappedCommand


MessageMaker = WrappedCommandFactory(MakeMessagesCommand, DEFAULT_MAKEMESSAGES_KWARGS)
MessageCompiler = WrappedCommandFactory(CompileMessagesCommand)



import hashlib
from functools import partial

import six


class CaseInsensitiveDict(dict):
    def __init__(self, obj = None, **kwargs):
        if obj is not None:
            if isinstance(obj, dict):
                for k, v in obj.items():
                    self[k] = v
            else:
                for k, v in obj:
                    self[k] = v

        for k, v in kwargs.items():
            self[k] = v

    def __setitem__(self, key, value):
        super(CaseInsensitiveDict, self).__setitem__(key.lower(), value)

    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(key.lower())

    def __delitem__(self, key):
        return super(CaseInsensitiveDict, self).__delitem__(key.lower())

    def __contains__(self, key):
        return super(CaseInsensitiveDict, self).__contains__(key.lower())

    def pop(self, key):
        return super(CaseInsensitiveDict, self).pop(key.lower())


class memoize(object):
    """
    Memoize the return parameters of a function.
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


if six.PY3:
    def map_apply(fn, iterable):
        return list(map(fn, iterable))
else:
    map_apply = map


def checksum(s):
    """
    Calculate the checksum of a string.
    Should eventually support files too.

    We use MD5 because S3 does.
    """
    return hashlib.md5(s).hexdigest()


def get_or_prompt(config, key, prompt_fn, *args, **kwargs):
    """
    :param config: The configuration object to get the value from
    :param key: The configuration key to retrieve
    :type key: str
    :param prompt_fn: The prompt function to use to prompt the value
    :param args: Extra arguments for the prompt function
    :param kwargs: Extra keyword arguments for hte prompt function
    """
    value = config.get(key)
    if value is None:
        value = prompt_fn(*args, **kwargs)
        config.set(key, value)
    return value



import os
import logging

from django.conf import settings
from django.template import Library
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe

logger = logging.getLogger(__name__)

register = Library()


def static(context, link_url):
    """
    Get the path for a static file in the Cactus build.
    We'll need this because paths can be rewritten with fingerprinting.
    """
    #TODO: Support URLS that don't start with `/static/`
    site = context['__CACTUS_SITE__']
    page = context['__CACTUS_CURRENT_PAGE__']

    url = site.get_url_for_static(link_url)

    if url is None:

        # For the static method we check if we need to add a prefix
        helper_keys = [
            "/static/" + link_url,
            "/static"  + link_url,
            "static/"  + link_url
        ]

        for helper_key in helper_keys:

            url_helper_key = site.get_url_for_static(helper_key)

            if url_helper_key is not None:
                return url_helper_key

        logger.warning('%s: static resource does not exist: %s', page.link_url, link_url)

        url = link_url

    return url


def url(context, link_url):
    """
    Get the path for a page in the Cactus build.
    We'll need this because paths can be rewritten with prettifying.
    """
    site = context['__CACTUS_SITE__']
    page = context['__CACTUS_CURRENT_PAGE__']

    url = site.get_url_for_page(link_url)

    if url is None:

        # See if we're trying to link to an /subdir/index.html with /subdir
        link_url_index = os.path.join(link_url, "index.html")
        url_link_url_index = site.get_url_for_page(link_url_index)

        if url_link_url_index is None:
            logger.warning('%s: page resource does not exist: %s', page.link_url, link_url)

        url = link_url

    locale = site.config.get("locale")
    if locale is not None and site.verb == site.VERB_BUILD:
        # prepend links with language directory
        url = u"/%s%s" % (site.config.get("locale"), url)

    if site.prettify_urls:
        return url.rsplit('index.html', 1)[0]

    return url


def config(context, key):
    """
    Get a value from the config by key
    """
    site = context['__CACTUS_SITE__']
    result = site.config.get(key)

    if result:
        return result

    return ""


def current_page(context):
    """
    Returns the current URL
    """
    page = context['__CACTUS_CURRENT_PAGE__']

    return page.final_url


def if_current_page(context, link_url, positive=True, negative=False):
    """
    Return one of the passed parameters if the URL passed is the current one.
    For consistency reasons, we use the link_url of the page.
    """
    page = context['__CACTUS_CURRENT_PAGE__']

    return positive if page.link_url == link_url else negative

@register.filter(is_safe=True)
def markdown(value, arg=''):
    """
    Runs Markdown over a given value, optionally using various
    extensions python-markdown supports.

    Syntax::

        {{ value|markdown2:"extension1_name,extension2_name..." }}

    To enable safe mode, which strips raw HTML and only returns HTML
    generated by actual Markdown syntax, pass "safe" as the first
    extension in the list.

    If the version of Markdown in use does not support extensions,
    they will be silently ignored.

    """
    try:
        import markdown2
    except ImportError:
        logging.warning("Markdown package not installed.")
        return force_str(value)
    else:
        def parse_extra(extra):
            if ':' not in extra:
                return (extra, {})
            name, values = extra.split(':', 1)
            values = dict((str(val.strip()), True) for val in values.split('|'))
            return (name.strip(), values)

        extras = (e.strip() for e in arg.split(','))
        extras = dict(parse_extra(e) for e in extras if e)

        if 'safe' in extras:
            del extras['safe']
            safe_mode = True
        else:
            safe_mode = False

        return mark_safe(markdown2.markdown(force_str(value), extras=extras, safe_mode=safe_mode))

register.simple_tag(takes_context=True)(static)
register.simple_tag(takes_context=True)(url)
register.simple_tag(takes_context=True)(config)
register.simple_tag(takes_context=True)(current_page)
register.simple_tag(takes_context=True)(if_current_page)



#coding:utf-8
import os
import time
import threading
import logging

from ctypes import *

from cactus.utils.filesystem import fileList
from cactus.utils.network import retry

from fsevents import Observer, Stream

class struct_timespec(Structure):
    _fields_ = [('tv_sec', c_long), ('tv_nsec', c_long)]

class struct_stat64(Structure):
    _fields_ = [
        ('st_dev', c_int32),
        ('st_mode', c_uint16),
        ('st_nlink', c_uint16),
        ('st_ino', c_uint64),
        ('st_uid', c_uint32),
        ('st_gid', c_uint32),
        ('st_rdev', c_int32),
        ('st_atimespec', struct_timespec),
        ('st_mtimespec', struct_timespec),
        ('st_ctimespec', struct_timespec),
        ('st_birthtimespec', struct_timespec),
        ('dont_care', c_uint64 * 8)
    ]

libc = CDLL('/usr/lib/libc.dylib')
stat64 = libc.stat64
stat64.argtypes = [c_char_p, POINTER(struct_stat64)]

# OS-X only function to get actual creation date
def get_creation_time(path):
    buf = struct_stat64()
    rv = stat64(path, pointer(buf))
    if rv != 0:
        raise OSError("Couldn't stat file %r" % path)
    return buf.st_birthtimespec.tv_sec


def createStream(real_path, link_path, callback):

    def cb(event):

        if not os.path.isdir(link_path):
            event.name = link_path
        else:
            translated_path = os.path.join(link_path, os.path.relpath(event.name, real_path))
            event.name = translated_path

        callback(event)

    return Stream(cb, real_path, file_events=True)


class FSEventsListener(object):
    def __init__(self, path, f, ignore = None):

        logging.info("Using FSEvents")

        self.path = path
        self.f = f
        self.ignore = ignore

        self.observer = Observer()
        self.observer.daemon = True

        self._streams = []
        self._streams.append(createStream(self.path, path, self._update))

        self._streamed_folders = [self.path]

        def add_stream(p):
            if p in self._streamed_folders:
                return
            self._streams.append(
                createStream(p, file_path, self._update))
            self._streamed_folders.append(p)

        # Start an extra listener for all symlinks
        for file_path in fileList(self.path, folders=True):
            if os.path.islink(file_path):
                if os.path.isdir(file_path):
                    add_stream(os.path.realpath(file_path))
                else:
                    add_stream(os.path.dirname(os.path.realpath(file_path)))

    def run(self):
        self.resume()
        self.observer.start()

    def pause(self):
        logging.debug("MacListener.PAUSE")

        for stream in self._streams:
            self.observer.unschedule(stream)

    def resume(self):
        logging.debug("MacListener.RESUME")

        for stream in self._streams:
            self.observer.schedule(stream)

    def stop():
        self.observer.stop()

    def _update(self, event):

        path = event.name

        if self.ignore and self.ignore(path):
            return

        logging.debug("MacListener.update %s", event)

        result = {
            'added': [],
            'deleted': [],
            'changed': [],
        }

        if os.path.exists(path):

            seconds_since_created = int(time.time()) - get_creation_time(os.path.realpath(path))

            if seconds_since_created < 1.0:
                result["added"].append(path)
            else:
                result["changed"].append(path)
        else:
            result["deleted"].append(path)

        self.f(result)



#coding:utf-8

from cactus.deployment.file import BaseFile
from cactus.utils.helpers import CaseInsensitiveDict


class CloudFilesFile(BaseFile):
    def remote_changed(self):
        obj = self.engine.bucket.get_object(self.url)
        #TODO: Headers
        return obj.etag != self.payload_checksum

    def get_headers(self):
        headers = CaseInsensitiveDict()
        for k in ("Cache-Control", "X-TTL"):
            headers[k] = 'max-age={0}'.format(self.cache_control)
        if self.content_encoding is not None:
            headers['Content-Encoding'] = self.content_encoding
        return headers

    def do_upload(self):
        obj = self.engine.bucket.store_object(self.url, self.payload(), content_type=self.content_type,
                                              etag=self.payload_checksum, content_encoding=self.content_encoding,)
        obj.set_metadata(self.get_headers())



import os

from six.moves import urllib, http_client

from cactus.utils.helpers import CaseInsensitiveDict


EXTERNAL_SCHEMES = ("//", "http://", "https://", "mailto:")


def getURLHeaders(url):
    url = urllib.parse.urlparse(url)

    conn = http_client.HTTPConnection(url.netloc)
    conn.request('HEAD', urllib.parse.quote(url.path))

    response = conn.getresponse()

    return CaseInsensitiveDict(response.getheaders())


def is_external(url):
    for scheme in EXTERNAL_SCHEMES:
        if url.startswith(scheme):
            return True
    return False


def path_to_url(path):
    """
    Convert a system path to an URL
    """
    return path.replace(os.sep, "/")


def URLHelperMixinFactory(class_name, property_name):

    inner_property_name = "_" + property_name

    def setter(self, value):
        setattr(self, inner_property_name, value)

    def getter(self):
        return path_to_url(getattr(self, inner_property_name))

    def deleter(self):
        delattr(self, inner_property_name)

    return type(class_name, (object,), {
        property_name: property(getter, setter, deleter)
    })


class ResourceURLHelperMixin(
    URLHelperMixinFactory("LinkURLHelperMixin", "link_url"),
    URLHelperMixinFactory("FinalURLHelperMixin", "final_url")
):
    """
    This helper lets us set the `link_url` and `final_url` for our resources and ensure that we get compliant URLS on
    all systems, including Windows where os.path is not a POSIX path.
    """



# #coding:utf-8
# import os
# import shutil
# import tempfile
# import keyring
# import keyring.backend
# import unittest
#
# from cactus.deployment.s3.auth import AWSCredentialsManager
# from cactus.config.file import ConfigFile
#
#
# class DummySite(object):
#     def __init__(self, config):
#         self.config = config
#
#
# class DummyEngine(object):
#     def __init__(self, site):
#         self.site = site
#
#
# class TestKeyring(keyring.backend.KeyringBackend):
#     def __init__(self):
#         self.password = ''
#
#     def supported(self):
#         return 0
#
#     def get_password(self, service, username):
#         return self.password
#
#     def set_password(self, service, username, password):
#         self.password = password
#         return 0
#
#     def delete_password(self, service, username):
#         self.password = None
#
#
# class CredentialsManagerTestCase(unittest.TestCase):
#     aws_access_key = "123"
#     aws_secret_key = "456"
#
#     def setUp(self):
#         self.test_dir = tempfile.mkdtemp()
#         self.path = os.path.join(self.test_dir, "conf1.json")
#         self.config = ConfigFile(self.path)
#
#         self.site = DummySite(self.config)
#         self.engine = DummyEngine(self.site)
#
#         self._initial_keyring = keyring.get_keyring()
#         self.keyring = TestKeyring()
#         keyring.set_keyring(self.keyring)
#
#         self.credentials_manager = AWSCredentialsManager(self.engine)
#
#     def tearDown(self):
#         shutil.rmtree(self.test_dir)
#         keyring.set_keyring(self._initial_keyring)
#
#     def test_read_config(self):
#         """
#         Test that credentials can be retrieved from the config
#         """
#         self.config.set("aws-access-key", self.aws_access_key)
#         self.keyring.set_password("aws", self.aws_access_key, self.aws_secret_key)
#
#         credentials = self.credentials_manager.get_credentials()
#         self.assertEqual(2, len(credentials))
#
#         username, password = credentials
#         self.assertEqual(self.aws_access_key, username)
#         self.assertEqual(self.aws_secret_key, password)
#
#     def test_write_config(self):
#         """
#         Test that credentials are persisted to a config file
#         """
#         self.credentials_manager.username = self.aws_access_key
#         self.credentials_manager.password = self.aws_secret_key
#         self.credentials_manager.save_credentials()
#
#         self.config.load()
#
#         self.assertEqual(self.aws_access_key, self.config.get("aws-access-key"))
#         self.assertEqual(self.aws_secret_key, self.keyring.get_password("aws", "123"))



from __future__ import print_function
import os

INFO = {
        'name': 'Version Updater',
        'description': 'Add a version to /versions.txt after each deploy'
}

# Set up extra django template tags

def templateTags():
    pass


# Build actions

# def preBuild(site):
#       print 'preBuild'
#
# def postBuild(site):
#       print 'postBuild'

# Build page actions

# def preBuildPage(site, path, context, data):
#       print 'preBuildPage', path
#       return context, data
#
# def postBuildPage(site, path):
#       print 'postBuildPage', path
#       pass


# Deploy actions

def preDeploy(site):

    # Add a deploy log at /versions.txt

    import urllib2
    import datetime
    import platform
    import codecs
    import getpass

    url = site.config.get('aws-bucket-website')
    data = u''

    # If this is the first deploy we don't have to fetch the old file
    if url:
        try:
            data = urllib2.urlopen('http://%s/versions.txt' % url, timeout=8.0).read() + u'\n'
        except:
            print("Could not fetch the previous versions.txt, skipping...")
            return

    data += u'\t'.join([datetime.datetime.now().isoformat(), platform.node(), getpass.getuser()])
    codecs.open(os.path.join(site.paths['build'], 'versions.txt'), 'w', 'utf8').write(data)

def postDeploy(site):
    pass



#coding:utf-8
import os
import logging

from cactus.deployment.file import BaseFile
from cactus.utils.filesystem import fileList
from cactus.utils.helpers import get_or_prompt, memoize, map_apply
from cactus.utils.parallel import multiMap, PARALLEL_DISABLED


logger = logging.getLogger(__name__)


class BaseDeploymentEngine(object):
    FileClass = BaseFile
    CredentialsManagerClass = None  #TODO: Define interface?

    config_bucket_name = None
    config_bucket_website = None

    _index_page = "index.html"
    _error_page = "error.html"

    _connection = None

    def __init__(self, site):
        """
        :param site: An instance of cactus.site.Site
        """
        self.site = site
        self.credentials_manager = self.CredentialsManagerClass(self)

    def deploy(self):
        self.configure()

        # Upload all files concurrently in a thread pool
        mapper = multiMap if self.site._parallel > PARALLEL_DISABLED else map_apply
        totalFiles = mapper(lambda p: p.upload(), self.files())

        return totalFiles

    def _ignore_file(self, path):

        if os.path.basename(path).startswith("."):
            return True

        # Special case for Finder Icon files
        if "\r" in os.path.basename(path):
            return True

        return False


    @memoize
    def files(self):
        """
        List of build files.
        """
        return [self.FileClass(self, file_path) for file_path in fileList(
            self.site.build_path, relative=True) if self._ignore_file(file_path) is False]

    def total_bytes(self):
        """
        Total size of files to be uploaded
        """
        return sum([f.total_bytes for f in self.files()])

    def total_bytes_uploaded(self):
        """
        Total size of files to be uploaded
        """
        return sum([f.total_bytes_uploaded for f in self.files()])

    def progress(self):
        """
        Progress of upload in percentage
        """
        total_bytes = float(self.total_bytes())
        total_bytes_uploaded = float(self.total_bytes_uploaded())

        if total_bytes == 0 or total_bytes_uploaded == 0:
            return 0.0

        return total_bytes_uploaded / total_bytes


    def get_connection(self):
        if self._connection is None:
            self._connection = self._create_connection()
        return self._connection

    def _create_connection(self):
        """
        Should return a Connection object
        """
        raise NotImplementedError()

    def get_bucket(self):
        """
        Should return a Bucket object, None if the bucket does not exist.
        """
        raise NotImplementedError()

    def create_bucket(self):
        """
        Should create and return a Bucket object.
        """
        raise NotImplementedError()

    def get_website_endpoint(self):
        """
        Should return the Website endpoint for the bucket.
        """
        #TODO: Normalize -- rackspace gives an URL, but Amazon gives a domain name
        raise NotImplementedError()

    def configure(self):
        """
        This is when the DeploymentEngine should configure itself to prepare for deployment

        :rtype: None
        """
        self.bucket_name = get_or_prompt(self.site.config, self.config_bucket_name, self.site.ui.prompt_normalized,
                                         "Enter the bucket name (e.g.: www.example.com)")
        #TODO: Validate this is not empty

        self.bucket = self.get_bucket()  #TODO: Catch auth errors

        #TODO: Make this all integrated and consistent!
        created = False
        if self.bucket is None:
            if self.site.ui.prompt_yes_no("Bucket does not exist. Create it?"):
                self.bucket = self.create_bucket()
                created = True
            else:
                return

        website_endpoint = self.get_website_endpoint()
        self.site.config.set(self.config_bucket_website, website_endpoint)

        self.site.config.write()
        self.credentials_manager.save_credentials()

        if created:
            logger.info('Bucket %s was created with website endpoint %s',  self.bucket_name, website_endpoint)

        logger.info("Bucket Name: %s", self.bucket_name)
        logger.info("Bucket Web Endpoint: %s", website_endpoint)



import os
import sys
import subprocess
import shutil

from distutils.sysconfig import get_python_lib
from setuptools import setup

PACKAGE_NAME = 'lava-cactus'
VERSION = "1.1.1"
SKELETON_FOLDERS = [
    'pages',
    'plugins',
    'static/css',
    'static/images',
    'static/js',
    'templates',
    'locale',
    '.github/workflows'
]
SKELETON_GLOB = ['skeleton/{0}/*'.format(folder) for folder in SKELETON_FOLDERS]

if "uninstall" in sys.argv:

    def run(command):
        try:
            return subprocess.check_output(command, shell=True).strip()
        except subprocess.CalledProcessError:
            pass


    cactusBinPath = run(f'which {PACKAGE_NAME}')
    cactusPackagePath = None

    for p in os.listdir(get_python_lib()):
        if p.lower().startswith(PACKAGE_NAME) and p.lower().endswith('.egg'):
            cactusPackagePath = os.path.join(get_python_lib(), p)

    if cactusBinPath and os.path.exists(cactusBinPath):
        sys.stdout.write('Removing cactus script at %s' % cactusBinPath)
        os.unlink(cactusBinPath)

    if cactusPackagePath and os.path.isdir(cactusPackagePath):
        sys.stdout.write('Removing cactus package at %s' % cactusPackagePath)
        shutil.rmtree(cactusPackagePath)

    sys.exit()

if "install" in sys.argv or "bdist_egg" in sys.argv:

    # Check if we have an old version of cactus installed
    p1 = '/usr/local/bin/cactus.py'
    p2 = '/usr/local/bin/cactus.pyc'

    if os.path.exists(p1) or os.path.exists(p2):
        sys.stdout.write("Error: you have an old version of Cactus installed, we need to move it:")
        if os.path.exists(p1):
            sys.stderr.write("  sudo rm %s" % p1)
        if os.path.exists(p2):
            sys.stderr.write("  sudo rm %s" % p2)
        sys.exit()


# From Django

def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join)
    in a platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


EXCLUDE_FROM_PACKAGES = ['cactus.skeleton', ]


def is_package(package_name):
    for pkg in EXCLUDE_FROM_PACKAGES:
        if package_name.startswith(pkg):
            return False
    return True


# Compile the list of packages available, because distutils doesn't have
# an easy way to do this.
packages, package_data = [], {}

root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)
cactus_dir = 'cactus'

for dirpath, dirnames, filenames in os.walk(cactus_dir):
    # Ignore PEP 3147 cache dirs and those whose names start with '.'
    dirnames[:] = [d for d in dirnames if not d.startswith('.') and d != '__pycache__']
    parts = fullsplit(dirpath)
    package_name = '.'.join(parts)
    if '__init__.py' in filenames and is_package(package_name):
        packages.append(package_name)
    elif filenames:
        relative_path = []
        while '.'.join(parts) not in packages:
            relative_path.append(parts.pop())
        relative_path.reverse()
        path = os.path.join(*relative_path)
        package_files = package_data.setdefault('.'.join(parts), [])
        package_files.extend([os.path.join(path, f) for f in filenames])


def find_requirements():
    # Find all requirements.VERSION.txt files that match (e.g. Python 2.6 matches
    # requirements.2.6.txt, requirements.2.txt, and requirments.txt).
    v = [str(x) for x in sys.version_info[:2]]
    requirements = []
    while True:
        reqs_file = '.'.join(["requirements"] + v + ["txt"])
        try:
            with open(os.path.join(root_dir, reqs_file)) as f:
                requirements.extend(f.readlines())
        except IOError:
            pass
        try:
            v.pop()
        except IndexError:
            break

    return requirements


def read(filename):
    with open(filename, 'r', encoding='utf-8') as file:
        return file.read()


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description="Static site generation and deployment.",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/quillcraftsman/lavacactus',
    author='quillcraftsman',
    author_email='quill@craftsman.lol',
    license='BSD',
    packages=packages,
    package_data=package_data,
    entry_points={
        'console_scripts': [
            'cactus = cactus.cli:cli_entrypoint',
        ],
    },
    install_requires=find_requirements(),
    tests_require=open(os.path.join(root_dir, "test_requirements.txt")).readlines(),
    zip_safe=False,
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: WWW/HTTP",
    ],
)



#coding:utf-8
import os
import io
import logging

from cactus import mime
from cactus.utils import ipc
from cactus.utils.file import compressString, fileSize
from cactus.utils.helpers import memoize, checksum
from cactus.utils.url import path_to_url


logger = logging.getLogger(__name__)


class BaseFile(object):
    DEFAULT_CACHE_EXPIRATION = 60 * 60 * 24 * 7  # One week
    MAX_CACHE_EXPIRATION = 60 * 60 * 24 * 365 # 1 Year (for cached)
    PROGRESS_MIN_SIZE = (1024 * 1024) / 2  # 521 kb

    def __init__(self, engine, path):
        self.engine = engine
        self.path = path

        self.force_refresh = False
        self._is_compressed = None

        self.total_bytes = len(self.payload())
        self.total_bytes_uploaded = 0

    def prepare(self):
        """
        Prepare the file for upload
        """
        payload = self.payload()  # Decide whether we'll compress or not.
        self.payload_checksum = checksum(payload)
        self.lastUpload = 0

    @property
    def url(self):
        """
        We must use this when deploying, otherwise the paths will be broken when
        deploying from Windows
        """
        return path_to_url(self.path)

    @memoize
    def data(self):
        with io.FileIO(os.path.join(self.engine.site.build_path, self.path), 'r') as f:
            return f.read()

    def payload(self):
        """
        The representation of the data that should be uploaded to the
        server. This might be compressed based on the content type and size.
        """
        if not hasattr(self, '_payload'):
            raw_data = self.data()

            if self.extension() in self.engine.site.compress_extensions:
                compressed_data = compressString(raw_data)

                if len(compressed_data) < len(raw_data):
                    self._is_compressed = True
                    self._payload = compressed_data
                    return self._payload

            self.is_compressed = False
            self._payload = raw_data

        return self._payload

    @property
    def is_compressed(self):
        assert self._is_compressed is not None, "Compression is not defined yet!"
        return self._is_compressed

    @is_compressed.setter
    def is_compressed(self, boolean):
        self._is_compressed = boolean

    @property
    def is_fingerprinted(self):
        """
        Lazy implementation (needs to be fixed); we'll re-fingerprint the file
        and check whether that happens to be in the filename.
        """
        #TODO: FixMe!
        return checksum(self.data()) in self.path

    def extension(self):
        return os.path.splitext(self.path)[1].strip('.').lower()

    @property
    def content_type(self):
        """
        Return the content type for this object
        """
        content_type = mime.guess(self.path)

        if not content_type:
            return None

        if content_type == "text/html":
            content_type = "{0}; charset=utf-8".format(content_type)

        return content_type

    def must_refresh(self):
        if self.force_refresh:
            return True

        return self.remote_changed()

    def upload(self):

        self.prepare()

        # Plugins may actually update those values afterwards
        self.cache_control = self.MAX_CACHE_EXPIRATION if self.is_fingerprinted else self.DEFAULT_CACHE_EXPIRATION
        self.content_encoding = 'gzip' if self.is_compressed else None
        self.content_length = len(self.payload())

        self.engine.site.plugin_manager.preDeployFile(self)

        remote_changed = self.remote_changed()

        if remote_changed:
            self.do_upload()

        self.total_bytes_uploaded = self.total_bytes

        op1 = '+' if remote_changed else '-'
        op2 = ' (%s compressed)' % (fileSize(len(self.payload()))) if self.is_compressed else ''

        # logger.warning("deploy.progress %s", self.engine.progress())

        ipc.signal("deploy.progress", {
            "progress": self.engine.progress(),
            "fileName": self.path
        })

        logger.info('%s %s - %s%s' % (op1, self.path, fileSize(len(self.data())), op2))

        return {'changed': remote_changed, 'size': len(self.payload())}

    def __repr__(self):
        return '<File: {0}>'.format(self.path)

    def remote_changed(self):
        """
        Did the file change when compared to the remote?
        :rtype: bool
        """
        raise NotImplementedError()

    def do_upload(self):
        """
        Actually upload the file to the remote
        """
        raise NotImplementedError()



#coding:utf-8
import unittest

from cactus import ui


class UITestCase(unittest.TestCase):
    def test_coerce_yes_no(self):
        self.assertEqual(True, ui._yes_no_coerce_fn("y"))
        self.assertEqual(True, ui._yes_no_coerce_fn("Y"))

        self.assertEqual(False, ui._yes_no_coerce_fn("n"))
        self.assertEqual(False, ui._yes_no_coerce_fn("N"))

        self.assertRaises(ui.InvalidInput, ui._yes_no_coerce_fn, "True")
        self.assertRaises(ui.InvalidInput, ui._yes_no_coerce_fn, "False")
        self.assertRaises(ui.InvalidInput, ui._yes_no_coerce_fn, "yes")
        self.assertRaises(ui.InvalidInput, ui._yes_no_coerce_fn, "no")

    def test_coerce_normalized(self):
        self.assertEqual("a", ui._normalized_coerce_fn("a "))
        self.assertEqual("a", ui._normalized_coerce_fn("A "))
        self.assertEqual("a", ui._normalized_coerce_fn(" A "))

    def test_coerce_url(self):
        self.assertEqual("http://www.example.com/", ui._url_coerce_fn("http://www.example.com/"))
        self.assertEqual("http://www.example.com/", ui._url_coerce_fn("http://www.EXAMPLE.com/"))
        self.assertEqual("http://www.example.com/", ui._url_coerce_fn("http://www.example.com"))

        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "")
        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "www.example.com")
        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "www.example.com  ")
        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "http://")
        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "/")
        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "http://www.example.com/somewhere/")
        self.assertRaises(ui.InvalidInput, ui._url_coerce_fn, "http://www.example.com/#hash")


# Disabled for now with the desktop app

# class InteractiveUITestCase(BaseTestCase):
#     def test_site_url_not_set(self):
#         class DummyUI(object):
#             def prompt_url(self, q):
#                 return "http://example.com"

#         site = Site(self.path, ui=DummyUI())
#         self.assertEqual(None, site.url)
#         site.build()
#         self.assertEqual("http://example.com", site.url)



#coding:utf-8

class VirtualPaths(object):
    def __init__(self, obj, mapping):
        self.obj = obj
        self.mapping = mapping

    def __getitem__(self, item):

        return getattr(self.obj, self.mapping[item])


class CompatibilityLayer(object):
    """
    Ensure backwards compatibility with older versions of Cactus.
    """
    mapping = {}

    @property
    def paths(self):
        return VirtualPaths(self, self.mapping)


class SiteCompatibilityLayer(CompatibilityLayer):
    mapping = {
        'build': 'build_path',
        'pages': 'page_path',
        'plugins': 'plugin_path',
        'templates': 'template_path',
        'static': 'static_path',
        'script': 'script_path',
    }


class PageCompatibilityLayer(CompatibilityLayer):
    mapping = {
        'full': 'full_source_path',
        'full-build': 'full_build_path',
    }

    @property
    def path(self):
        return self.source_path


class StaticCompatibilityLayer(CompatibilityLayer):
    mapping = {
        'full': 'full_source_path',
        'full-build': 'full_build_path',
    }



#coding:utf-8
from cactus.deployment.auth import BaseKeyringCredentialsManager


class AWSCredentialsManager(BaseKeyringCredentialsManager):
    _username_config_entry = "aws-access-key"
    _username_display_name = "Amazon Access Key ID"
    _password_display_name = "Amazon Secret Access Key"
    _keyring_service = "aws"  #Would break backwards compatibility



#coding:utf-8
import os

from cactus.tests import SiteTestCase


class TestPageTags(SiteTestCase):
    """
    {% current_page %} should return the URL the user is currently seeing.
    {% url 'page.html' %} should return the link to a page
    {% if_current_page 'page.html' 1 0 %} should return 1 if we're on this page, 0 otherwise
    """
    def setUp(self):
        super(TestPageTags, self).setUp()

        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write("{% current_page %}")

        with open(os.path.join(self.site.page_path, "link.html"), "w") as f:
            f.write("{% url '/test.html' %}")

        with open(os.path.join(self.site.page_path, "if-yes.html"), "w") as f:
            f.write("{% if_current_page '/if-yes.html' 1 0 %}")

        with open(os.path.join(self.site.page_path, "if-no.html"), "w") as f:
            f.write("{% if_current_page '/if-yes.html' 1 0 %}")

    def test_regular_urls(self):
        self.site.build()

        with open(os.path.join(self.site.build_path, "test.html")) as f:
            self.assertEqual(f.read(), "/test.html")

        with open(os.path.join(self.site.build_path, "link.html")) as f:
            self.assertEqual(f.read(), "/test.html")

        with open(os.path.join(self.site.build_path, "if-yes.html")) as f:
            self.assertEqual(f.read(), "1")

        with open(os.path.join(self.site.build_path, "if-no.html")) as f:
            self.assertEqual(f.read(), "0")

    def test_pretty_urls(self):
        self.site.prettify_urls = True  #TOOD: should be read at runtime
        self.site.build()

        with open(os.path.join(self.site.build_path, "test", "index.html")) as f:
            self.assertEqual(f.read(), "/test/")

        with open(os.path.join(self.site.build_path, "link", "index.html")) as f:
            self.assertEqual(f.read(), "/test/")

        with open(os.path.join(self.site.build_path, "if-yes", "index.html")) as f:
            self.assertEqual(f.read(), "1")

        with open(os.path.join(self.site.build_path, "if-no", "index.html")) as f:
            self.assertEqual(f.read(), "0")


class TestStatic(SiteTestCase):
    """
    {% static %} should convert from a link_url to a final_url
    """
    def setUp(self):
        super(TestStatic, self).setUp()

        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write("{% static '/static/js/main.js' %}")

    def test_no_fingerprints(self):
        self.site.build()

        with(open(os.path.join(self.site.build_path, "test.html"))) as f:
            self.assertEqual(f.read(), "/static/js/main.js")

    def test_fingerprint(self):
        self.site.build()

        expected = None

        for static in self.site.static():
            if static.src_filename == "main.js":
                expected = static.final_url

        self.assertIsNotNone(expected)  # Was the JS built?

        with open(os.path.join(self.site.build_path, "test.html")) as f:
            self.assertEqual(f.read(), expected)

class TestStaticLookup(SiteTestCase):
    """
    {% static %} should auto append static
    """
    # def setUp(self):
    #     super(TestStaticLookup, self).setUp()

    def test_prepend_exists(self):

        with open(os.path.join(self.site.static_path, "exists.js"), "w") as f:
            f.write("hello")

        with open(os.path.join(self.site.page_path, "test1.html"), "w") as f:
            f.write("{% static '/exists.js' %}")

        with open(os.path.join(self.site.page_path, "test2.html"), "w") as f:
            f.write("{% static 'exists.js' %}")

        self.site.build()

        with(open(os.path.join(self.site.build_path, "test1.html"))) as f:
            self.assertEqual(f.read(), "/static/exists.js")

        with(open(os.path.join(self.site.build_path, "test2.html"))) as f:
            self.assertEqual(f.read(), "/static/exists.js")

    def test_prepend_nonexists(self):

        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write("{% static '/notexists.js' %}")

        self.site.build()

        with(open(os.path.join(self.site.build_path, "test.html"))) as f:
            self.assertEqual(f.read(), "/notexists.js")

class TestStaticRelative(SiteTestCase):

    def doit(self, static_subdir, page_subdir, page_contents, output):

        full_static_subdir = os.path.join(self.site.static_path, static_subdir)
        full_page_subdir = os.path.join(self.site.page_path, page_subdir)

        try: os.makedirs(full_static_subdir)
        except: pass

        try: os.makedirs(full_page_subdir)
        except: pass

        # Write the static file
        with open(os.path.join(full_static_subdir, "test.js"), "w") as f:
            f.write("hello")

        # Write the page with contents
        with open(os.path.join(full_page_subdir, "test.html"), "w") as f:
            f.write(page_contents)

        self.site.build()

        with(open(os.path.join(self.site.build_path, page_subdir, "test.html"))) as f:
            self.assertEqual(f.read(), output)

    def test_simple(self):
        self.doit("", "", "{% static 'static/test.js' %}", "static/test.js")

    # I'm not sure if we should do this at all

    # def test_relative(self):
    #     self.doit("", "docs", "{% static 'static/test.js' %}", "../static/test.js")




class TestMarkdown(SiteTestCase):

    def test_tags(self):

        for level in [1, 2, 3, 4, 5]:

            levelStr = "#" * level

            with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
                f.write("{% filter markdown %}" + levelStr + " Hello{% endfilter %}")

            self.site.build()

            with open(os.path.join(self.site.build_path, "test.html")) as f:
                self.assertEqual(f.read(), "<h%s>Hello</h%s>\n" % (level, level))

class TestConfig(SiteTestCase):

    def setUp(self):
        super(TestConfig, self).setUp()

        self.site.config.set("test-value", "yep")

    def test_config_value(self):

        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write("{% config 'test-value' %}")

        self.site.build()

        with open(os.path.join(self.site.build_path, "test.html")) as f:
            self.assertEqual(f.read(), "yep")

    def test_no_config_value(self):

        with open(os.path.join(self.site.page_path, "test.html"), "w") as f:
            f.write("{% config 'unknown' %}")

        self.site.build()

        with open(os.path.join(self.site.build_path, "test.html")) as f:
            self.assertEqual(f.read(), "")



#coding:utf-8
import os
import sys
import imp
import logging

from cactus.plugin import defaults
from cactus.utils.filesystem import fileList


logger = logging.getLogger(__name__)


class BasePluginsLoader(object):
    def load(self):
        raise NotImplementedError("Subclasses must implement load")

    def _initialize_plugin(self, plugin):
        """
        :param plugin: A plugin to initialize.
        :returns: An initialized plugin with all default methods set.
        """
        # Load default attributes
        for attr in defaults.DEFAULTS + ['ORDER']:
            if not hasattr(plugin, attr):
                setattr(plugin, attr, getattr(defaults, attr))

        # Name the plugin
        if not hasattr(plugin, "plugin_name"):
            if hasattr(plugin, "__name__"):
                plugin.plugin_name = plugin.__name__
            elif hasattr(plugin, "__class__"):
                plugin.plugin_name = plugin.__class__.__name__
            else:
                plugin.plugin_name = "anonymous"


class ObjectsPluginLoader(BasePluginsLoader):
    """
    Loads the plugins objects passed to this loader.
    """
    def __init__(self, plugins):
        """
        :param plugins: The list of plugins this loader should load.
        """
        self.plugins = plugins

    def load(self):
        """
        :returns: The list of plugins loaded by this loader.
        """
        plugins = []

        # Load cactus internal plugins
        for builtin_plugin in self.plugins:
            self._initialize_plugin(builtin_plugin)
            plugins.append(builtin_plugin)

        return plugins


class CustomPluginsLoader(BasePluginsLoader):
    """
    Loads all the plugins found at the path passed.
    """

    def __init__(self, plugin_path):
        """
        :param plugin_path: The path where the plugins should be loaded from.
        """
        self.plugin_path = plugin_path

    def load(self):
        """
        :returns: The list of plugins loaded by this loader.
        """
        plugins = []

        # Load user plugins
        for plugin_path in fileList(self.plugin_path):
            if self._is_plugin_path(plugin_path):
                custom_plugin = self._load_plugin_path(plugin_path)
                if custom_plugin:
                    self._initialize_plugin(custom_plugin)
                    plugins.append(custom_plugin)


        return plugins

    def _is_plugin_path(self, plugin_path):
        """
        :param plugin_path: A path where to look for a plugin.
        :returns: Whether this path looks like an enabled plugin.
        """
        if not plugin_path.endswith('.py'):
            return False

        if 'disabled' in plugin_path:
            return False

        return True

    def _load_plugin_path(self, plugin_path):
        """
        :param plugin_path: A path to load as a plugin.
        :returns: A plugin module.
        """
        module_name = "plugin_{0}".format(os.path.splitext(os.path.basename(plugin_path))[0])

        try:
            return imp.load_source(module_name, plugin_path)
        except Exception as e:
            logger.warning('Could not load plugin at path %s: %s' % (plugin_path, e))
            return None

            # sys.exit()



#coding:utf-8
from cactus.deployment.auth import BaseKeyringCredentialsManager


class CloudFilesCredentialsManager(BaseKeyringCredentialsManager):
    _username_config_entry = "cloudfiles-username"
    _password_display_name = "API Key"
    _keyring_service = "cactus/cloudfiles"



#coding:utf-8
import logging


logger = logging.getLogger(__name__)


class ConfigFallback(object):
    """
    A transient fallback config.
    This config does not store anything to the filesystem.
    """
    def __init__(self):
        self.cnf = {}

    @property
    def path(self):
        return "@fallback"

    def get(self, key, default=None):
        return self.cnf.get(key, default)

    def set(self, key, value):
        self.cnf[key] = value

    def has_key(self, key):
        return key in self.cnf

    def write(self):
        if self.cnf:
            logger.warning("Using config fallback, discarding config values: [%s]", ', '.join(self.cnf.keys()))



import os
import tempfile
import threading
import unittest

from cactus.listener import PollingListener
from cactus.tests.compat import has_symlink

try:
    from cactus.listener import FSEventsListener
except ImportError:
    FSEventsListener = None


def sleep(s):
    # time.sleep(s)
    event = threading.Event()
    event.wait(timeout=s)


class PollingListenerTest(unittest.TestCase):

    def setUp(self):
        self._callbacks = []
        self._callback_count = 0
        self.lock = threading.Lock()


    def _callback(self, event):
        with self.lock:
            self._callbacks.append(event)

    @property
    def callbacks(self):
        with self.lock:
            return self._callbacks

    def wait(self):

        time_total = 0
        time_increment = 0.03
        time_max = 3.0

        while True:

            # print time_total

            self.assertEqual(time_total < time_max, True, "Timeout")

            if len(self.callbacks) == self._callback_count + 1:
                self._callback_count += 1
                return

            sleep(time_increment)
            time_total += time_increment

    def create_listener(self, path):
        return PollingListener(path, self._callback)

    @unittest.skipUnless(has_symlink, "No symlink support")
    def testSymlinkFolder(self):

        path1 = os.path.realpath(os.path.join(tempfile.mkdtemp(), "watched"))
        path2 = os.path.realpath(os.path.join(tempfile.mkdtemp(), "linked"))

        os.mkdir(path1)
        os.mkdir(path2)

        path_link = os.path.join(path1, "path2")
        path_link_file = os.path.join(path2, "file.js")
        path_link_file_original = os.path.join(path_link, "file.js")

        os.symlink(path2, path_link)

        with open(path_link_file, "w") as f:
            f.write("hello1")

        self.listener = self.create_listener(path1)
        self.listener.run()

        sleep(1)

        with open(path_link_file, "w") as f:
            f.write("hello2")

        self.wait()

        self.assertEqual(len(self.callbacks), 1)
        self.assertEqual(self.callbacks[0]["changed"], [path_link_file_original])

    def testSimpleFile(self):

        path_watch = os.path.realpath(os.path.join(tempfile.mkdtemp(), "watched"))
        path_file = os.path.join(path_watch, "file.js")

        os.mkdir(path_watch)

        self.listener = self.create_listener(path_watch)
        self.listener.run()

        sleep(1)

        with open(path_file, "w") as f:
            f.write("hello1")

        self.wait()

        self.assertEqual(len(self.callbacks), 1)
        self.assertEqual(self.callbacks[0]["added"], [path_file])

        os.remove(path_file)

        self.wait()

        self.assertEqual(len(self.callbacks), 2)
        self.assertEqual(self.callbacks[1]["deleted"], [path_file])

    @unittest.skipUnless(has_symlink, "No symlink support")
    def testSymlinkFile(self):

        path_watch = os.path.realpath(os.path.join(tempfile.mkdtemp(), "watched"))
        path_linked = os.path.realpath(os.path.join(tempfile.mkdtemp(), "linked"))

        os.mkdir(path_watch)
        os.mkdir(path_linked)

        file_actual1 = os.path.join(path_linked, "framer.js")
        file_actual2 = os.path.join(path_linked, "framer.js")
        file_link = os.path.join(path_watch, "framer.js")

        with open(file_actual1, "w") as f:
            f.write("hello1")

        with open(file_actual2, "w") as f:
            f.write("hello2")

        os.symlink(file_actual1, file_link)

        self.listener = self.create_listener(path_watch)
        self.listener.run()

        if hasattr(self.listener, "streamed_folders"):
            self.assertEqual(len(self.listener.streamed_folders), 2)

        sleep(1)

        # Change the linked file
        with open(file_actual1, "w") as f:
            f.write("hello1")

        self.wait()

        self.assertEqual(len(self.callbacks), 1)
        self.assertEqual(self.callbacks[0]["changed"], [file_link])

        # sleep(1)

        # Now change a file in the same dir as the link, this should not do anything
        with open(file_actual2, "w") as f:
            f.write("hello333")

        sleep(1)

        with open(file_actual1, "w") as f:
            f.write("hello3223")

        self.wait()

        self.assertEqual(len(self.callbacks), 2)
        self.assertEqual(self.callbacks[0]["changed"], [file_link])

if FSEventsListener:

    class FSEventsListenerTest(PollingListener):

        def create_listener(self, path):
            return FSEventsListener(path, self._callback)



#coding:utf-8
from cactus.tests.deployment import DummyUI, DummySite, DummyDeploymentEngine, BaseDeploymentTestCase


class BucketCreateTestCase(BaseDeploymentTestCase):
    def setUp(self):
        super(BucketCreateTestCase, self).setUp()
        self.ui = DummyUI()
        self.site = DummySite(self.test_dir, self.ui)
        self.engine = DummyDeploymentEngine(self.site)

    def test_bucket_does_not_exist(self):
        """
        Test that we create buckets that don't exist
        """
        self.assertEqual(0, self.engine.create_bucket_calls)
        self.assertEqual(0, self.ui.asked_create)

        self.engine.configure()

        self.assertEqual(1, self.engine.get_bucket_calls)
        self.assertEqual(1, self.engine.create_bucket_calls)
        self.assertEqual(1, self.ui.asked_create)

    def test_bucket_exists(self):
        """
        Test that we don't attempt to re-create buckets that exist
        """
        self.engine.create_bucket_calls = 1

        self.engine.configure()

        self.assertEqual(1, self.engine.create_bucket_calls)
        self.assertEqual(1, self.engine.get_bucket_calls)
        self.assertEqual(0, self.ui.asked_create)



#! /usr/bin/python
# Dispatch - synchronize two folders
from __future__ import print_function

import os
import filecmp
import shutil

from stat import *


class Dispatch:
    ''' This class represents a synchronization object '''

    def __init__(self, name=''):
        self.name = name
        self.node_list = []
        self.file_copied_count = 0
        self.folder_copied_count = 0

    def add_node(self, node):
        self.node_list.append(node)

    def compare_nodes(self):
        ''' This method takes the nodes in the node_list and compares them '''
        # For each node in the list
        for node in self.node_list:
            # If the list has another item after it, compare them
            if self.node_list.index(node) < len(self.node_list) - 1:
                node2 = self.node_list[self.node_list.index(node) + 1]
                print('\nComparing Node ' + str(self.node_list.index(node)) + ' and Node ' + str(self.node_list.index(node) + 1) + ':')
                # Passes the two root directories of the nodes to the recursive _compare_directories.
                self._compare_directories(node.root_path, node2.root_path)

    def _compare_directories(self, left, right):
        ''' This method compares directories. If there is a common directory, the
            algorithm must compare what is inside of the directory by calling this
            recursively.
        '''
        comparison = filecmp.dircmp(left, right)
        if comparison.common_dirs:
            for d in comparison.common_dirs:
                self._compare_directories(os.path.join(left, d), os.path.join(right, d))
        if comparison.left_only:
            self._copy(comparison.left_only, left, right)
        if comparison.right_only:
            self._copy(comparison.right_only, right, left)
        left_newer = []
        right_newer = []
        if comparison.diff_files:
            for d in comparison.diff_files:
                l_modified = os.stat(os.path.join(left, d)).st_mtime
                r_modified = os.stat(os.path.join(right, d)).st_mtime
                if l_modified > r_modified:
                    left_newer.append(d)
                else:
                    right_newer.append(d)
        self._copy(left_newer, left, right)
        self._copy(right_newer, right, left)

    def _copy(self, file_list, src, dest):
        ''' This method copies a list of files from a source node to a destination node '''
        for f in file_list:
            srcpath = os.path.join(src, os.path.basename(f))
            if os.path.isdir(srcpath):
                shutil.copytree(srcpath, os.path.join(dest, os.path.basename(f)))
                self.folder_copied_count = self.folder_copied_count + 1
                print('Copied directory \"' + os.path.basename(srcpath) + '\" from \"' + os.path.dirname(srcpath) + '\" to \"' + dest + '\"')
            else:
                shutil.copy2(srcpath, dest)
                self.file_copied_count = self.file_copied_count + 1
                print('Copied \"' + os.path.basename(srcpath) + '\" from \"' + os.path.dirname(srcpath) + '\" to \"' + dest + '\"')


class Node:
    ''' This class represents a node in a dispatch synchronization '''
    def __init__(self, path, name=''):
        self.name = name
        self.root_path = os.path.abspath(path)
        self.file_list = os.listdir(self.root_path)


def syncFiles(src, dst):
    dispatch = Dispatch('dispatch')
    dispatch.add_node(Node(src, 'node1'))
    dispatch.add_node(Node(dst, 'node2'))
    dispatch.compare_nodes()
